/* tclgpgme.c - Tcl interface to GPGME */

/*
 * TBD:
 *
 * - asynchronous functions, i.e.,
 *   gpgme_op_{encrypt,encrypt_sign,decrypt,descript_verify,sign,verify}_start
 *   gpgme_{get,set}_io_cbs
 *
 * - keyring functions, i.e.,
 *   gpgme_op_{import,export,genkey,delete,edit}*
 *
 * - test trustitem stuff
 *
 */

#include <tcl.h>
#include <stdio.h>
#include <stdlib.h>

#include <gpgme.h>


/* common datastructures */


struct kv_pair {
  char    *p_key;
  unsigned p_value;
};

typedef struct kv_pair kv_pair_t;


struct oc_pair {
    char           *p_key;
    Tcl_ObjCmdProc *p_proc;
};

typedef struct oc_pair oc_pair_t;


struct gpgme_common {
    Tcl_Interp      *gc_interp;
    Tcl_Command      gc_token;

    oc_pair_t       *gc_dispatch;
};

typedef struct gpgme_common gpgme_common_t;
#define gd_interp       gd_common.gc_interp
#define gd_token        gd_common.gc_token
#define gd_dispatch     gd_common.gc_dispatch

struct gpgme_ctx {
    gpgme_common_t   gd_common;

    unsigned long    gd_flags;
#define GD_KEYLIST      0x01            /* doing a keylist */
#define GD_KEYEOF       0x02
#define GD_TRUSTLIST    0x02            /* doing a trustlist */
#define GD_TRUSTEOF     0x04

    GpgmeCtx         gd_context;

    Tcl_Obj         *gd_passphrase_cb;
    char            *gd_passphrase;

    Tcl_Obj         *gd_progress_cb;

    Tcl_Obj         *gd_idle_cb;

    Tcl_HashTable    gd_keyTable;

    Tcl_HashTable    gd_trustTable;
};

typedef struct gpgme_ctx gpgme_ctx_t;


struct gpgme_recip {
    gpgme_common_t   gd_common;
    char             gd_name[BUFSIZ];

    GpgmeRecipients  gd_recipients;
};

typedef struct gpgme_recip gpgme_recip_t;


/* static variables */

static
Tcl_HashTable   ctxTable;      /* GpgmeCtx        -> gpgme_ctx_t*   */

static
Tcl_HashTable   recipTable;    /* GpgmeRecipients -> gpgme_recip_t* */

static
Tcl_HashTable   recipXTable;   /* char *          -> gpgme_recip_t* */


/* enum management */


static kv_pair_t dt_pairs[] = {
    { "none", GPGME_DATA_TYPE_NONE },
    { "mem" , GPGME_DATA_TYPE_MEM  },
    { "fd"  , GPGME_DATA_TYPE_FD   },
    { "file", GPGME_DATA_TYPE_FILE },
    { "cb"  , GPGME_DATA_TYPE_CB   },

    { NULL  , 0                    }
};

static kv_pair_t ss_pairs[] = {
    { "none" ,      GPGME_SIG_STAT_NONE        },
    { "good" ,      GPGME_SIG_STAT_GOOD        },
    { "bad"  ,      GPGME_SIG_STAT_BAD         },
    { "nokey",      GPGME_SIG_STAT_NOKEY       },
    { "nosig",      GPGME_SIG_STAT_NOSIG       },
    { "error",      GPGME_SIG_STAT_ERROR       },
    { "diff" ,      GPGME_SIG_STAT_DIFF        },
    { "expired",    GPGME_SIG_STAT_GOOD_EXP    },
    { "expiredkey", GPGME_SIG_STAT_GOOD_EXPKEY },

    { NULL   ,      0                          }
};

static kv_pair_t ss_flags[] = {
    { "valid"            , GPGME_SIGSUM_VALID       },
    { "green"            , GPGME_SIGSUM_GREEN       },
    { "red"              , GPGME_SIGSUM_RED         },
    { "key-revoked"      , GPGME_SIGSUM_KEY_REVOKED },
    { "key-expired"      , GPGME_SIGSUM_KEY_EXPIRED },
    { "signature-expired", GPGME_SIGSUM_SIG_EXPIRED },
    { "key-missing"      , GPGME_SIGSUM_KEY_MISSING },
    { "crl-missing"      , GPGME_SIGSUM_CRL_MISSING },
    { "crl-too-old"      , GPGME_SIGSUM_CRL_TOO_OLD },
    { "bad-policy"       , GPGME_SIGSUM_BAD_POLICY  },
    { "sys-error"        , GPGME_SIGSUM_SYS_ERROR   },

    { NULL               , 0                        }
};

static kv_pair_t sm_pairs[] = {
    { "normal", GPGME_SIG_MODE_NORMAL },
    { "detach", GPGME_SIG_MODE_DETACH },
    { "clear" , GPGME_SIG_MODE_CLEAR  },

    { NULL    , 0                     }
};

static kv_pair_t attr_pairs[] = {
    { "keyid"            , GPGME_ATTR_KEYID        },
    { "fingerprint"      , GPGME_ATTR_FPR          },
    { "algorithm"        , GPGME_ATTR_ALGO         },
    { "length"           , GPGME_ATTR_LEN          },
    { "created"          , GPGME_ATTR_CREATED      },
    { "expires"          , GPGME_ATTR_EXPIRE       },
    { "owner-trust"      , GPGME_ATTR_OTRUST       },
    { "userid"           , GPGME_ATTR_USERID       },
    { "name"             , GPGME_ATTR_NAME         },
    { "email"            , GPGME_ATTR_EMAIL        },
    { "comment"          , GPGME_ATTR_COMMENT      },
    { "validity"         , GPGME_ATTR_VALIDITY     },
    { "level"            , GPGME_ATTR_LEVEL        },
    { "type"             , GPGME_ATTR_TYPE         },
    { "is-secret"        , GPGME_ATTR_IS_SECRET    },
    { "key-revoked"      , GPGME_ATTR_KEY_REVOKED  },
    { "key-invalid"      , GPGME_ATTR_KEY_INVALID  },
    { "uid-revoked"      , GPGME_ATTR_UID_REVOKED  },
    { "uid-invalid"      , GPGME_ATTR_UID_INVALID  },
    { "key-capability"   , GPGME_ATTR_KEY_CAPS     },
    { "key-expired"      , GPGME_ATTR_KEY_EXPIRED  },
    { "key-disabled"     , GPGME_ATTR_KEY_DISABLED },
    { "serial"           , GPGME_ATTR_SERIAL       },
    { "issuer"           , GPGME_ATTR_ISSUER       },
    { "chainid"          , GPGME_ATTR_CHAINID      },
    { "signature-status" , GPGME_ATTR_SIG_STATUS   },
    { "error-token"      , GPGME_ATTR_ERRTOK       },
    { "signature-summary", GPGME_ATTR_SIG_SUMMARY  },

    { NULL            , 0                       }
};

static kv_pair_t validity_pairs[] = {
    { "unknown"  , GPGME_VALIDITY_UNKNOWN   },
    { "undefined", GPGME_VALIDITY_UNDEFINED },
    { "never"    , GPGME_VALIDITY_NEVER     },
    { "marginal" , GPGME_VALIDITY_MARGINAL  },
    { "full"     , GPGME_VALIDITY_FULL      },
    { "ultimate" , GPGME_VALIDITY_ULTIMATE  },

    { NULL       , 0                        }
};

static kv_pair_t capability_flags[] = {
    { "encrypt", 1<<2 },
    { "sign"   , 1<<1 },
    { "certify", 1<<0 },

    { NULL     , 0    }
};

static kv_pair_t protocol_pairs[] = {
    { "openpgp", GPGME_PROTOCOL_OpenPGP }, 
    { "cms"    , GPGME_PROTOCOL_CMS     },
    { "auto"   , GPGME_PROTOCOL_AUTO    },

    { NULL     , 0                      }
};


static kv_pair_t keylist_flags[] = {
    { "local" , GPGME_KEYLIST_MODE_LOCAL  },
    { "extern", GPGME_KEYLIST_MODE_EXTERN },
    { "sigs"  , GPGME_KEYLIST_MODE_SIGS   },

    { NULL    , 0                         }
};

static kv_pair_t ss_attrs[] = {
    { "fingerprint", GPGME_ATTR_FPR          },
    { "error-token", GPGME_ATTR_ERRTOK       },
    { "created"    , GPGME_ATTR_CREATED      },
    { "expires"    , GPGME_ATTR_EXPIRE       },
    { "validity"   , GPGME_ATTR_VALIDITY     },
    { "status"     , GPGME_ATTR_SIG_STATUS   },
    { "summary"    , GPGME_ATTR_SIG_SUMMARY  },

    { NULL         , 0                       }
};

static kv_pair_t type_pairs[] = {
    { "userid", 2 },
    { "key"   , 1 },

    { NULL    , 0 }
};


/* args management */

/* avoid a bug in Tcl's index caching code... */

static int
tcl_GetIndexFromObjStruct (Tcl_Interp *interp,
                           Tcl_Obj    *objPtr,
                           char      **tablePtr,
                           int         offset,
                           char       *msg,
                           int         flags,
                           int        *indexPtr) {
    Tcl_GetCharLength (objPtr);

    return Tcl_GetIndexFromObjStruct (interp, objPtr, tablePtr, offset, msg,
                                      flags, indexPtr);
}


static int
crack_args (Tcl_Interp *interp,
            int         objc,
            Tcl_Obj    *CONST objv[],
            kv_pair_t  *switches,
            int         optional,
            Tcl_Obj    *args[]) {
    int     i,
            result;

    Tcl_ResetResult (interp);
    for (objc--, objv++; objc > 0; objc -= 2, objv += 2) {
        int        offset;
        kv_pair_t *p;

        if ((result = tcl_GetIndexFromObjStruct (interp, objv[0],
                                                 (char **) switches,
                                                 sizeof *switches, "switch",
                                                 0, &offset)) != TCL_OK)
            return result;
        p = switches + offset;

        if (objc < 2) {
            Tcl_AppendResult (interp, "missing argument to \"",
                              Tcl_GetString (objv[0]), "\"", NULL);
            return TCL_ERROR;
        }
        args[p -> p_value] = objv[1];   
    }

    for (i = 0; i < optional; i++)
        if (!args[i]) {
            Tcl_AppendResult (interp, switches[i].p_key,
                              " switch must be provided", NULL);
            return TCL_ERROR;
        }

    return TCL_OK;
}


#if     0
static int
t2c_usage (Tcl_Interp *interp,
           Tcl_Obj    *objPtr,
           kv_pair_t   pairs[],
           unsigned    optional,
           char       *operation,
           kv_pair_t   flags[]) {
    kv_pair_t *p;

    Tcl_ResetResult (interp);
    Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objPtr), NULL);
    for (p = pairs; p -> p_key; p++) {
        char     buffer[BUFSIZ],
                  *cp,
                  *v;
        kv_pair_t *f;

        if (((v = operation) != NULL)
                && Tcl_StringCaseMatch ("-operation", p -> p_key, 0)) {}
        else if ((flags != NULL)
                     && Tcl_StringCaseMatch ("-flags", p -> p_key, 0)) {
            cp = buffer;
            for (f = flags, v = "{"; f -> p_key; f++, v = " ") {
                sprintf (cp, "%s%s", v, f -> p_key);
                cp += strlen (cp);
            }
            strcpy (cp, "}");
            v = buffer;
        } else
            v = "...";
        
        Tcl_AppendResult (interp, p -> p_value >= optional ? " ?" : " ",
                          p -> p_key, " ", v,
                          p -> p_value >= optional ? "?" : "", NULL);
    }

    return TCL_ERROR;
}
#endif


static Tcl_Obj *
c2t_int (Tcl_Interp   *interp,
         kv_pair_t     pairs[],
         int           value) {
    kv_pair_t *p;  

    for (p = pairs; p -> p_key; p++)
        if (p -> p_value == value)
            return Tcl_NewStringObj (p -> p_key, -1);

    return Tcl_NewIntObj (value);
}


static int
t2c_flags (Tcl_Interp *interp,
           Tcl_Obj    *objPtr,
           kv_pair_t   pairs[],
           unsigned   *flags) {
    int      i,
             nelem,
             result;

    *flags = 0;
    if (!objPtr)
        return TCL_OK;

    if ((result = Tcl_ListObjLength (interp, objPtr, &nelem)) != TCL_OK)
        return result;

    for (i = 0; i < nelem; i++) {
        int        offset;
        kv_pair_t *p;
        Tcl_Obj   *elemPtr;

        if ((result = Tcl_ListObjIndex (interp, objPtr, i, &elemPtr))
                != TCL_OK)
            return result;

        if ((result = tcl_GetIndexFromObjStruct (interp, elemPtr,
                                                 (char **) pairs,
                                                 sizeof *pairs,
                                                 "flag", 0, &offset))
                != TCL_OK)
            return result;
        p = pairs + offset;

        *flags |= p -> p_value;
    }

    return TCL_OK;
}


static Tcl_Obj *
c2t_flags (Tcl_Interp   *interp,
           kv_pair_t     pairs[],
           unsigned long flags) {
    unsigned long ul = flags;
    kv_pair_t    *p;
    Tcl_Obj      *resultPtr;

    for (p = pairs; p -> p_key; p++)
        ul &= ~p -> p_value;
    if (ul)
        return Tcl_NewLongObj (flags);

    resultPtr = Tcl_NewObj ();

    for (p = pairs; p -> p_key; p++)
        if (p -> p_value & flags)
            Tcl_ListObjAppendElement (interp, resultPtr,
                                      Tcl_NewStringObj (p -> p_key, -1));

    return resultPtr;
}

/* result management */

static int
c2t_result (Tcl_Interp *interp,
            char       *fnx,
            int         result) {
    char   *cp,
            buffer[BUFSIZ];

    Tcl_ResetResult (interp);
    if (result == GPGME_No_Error)
        return TCL_OK;

    Tcl_SetResult (interp, cp = (char *) gpgme_strerror (result),
                   TCL_VOLATILE);

    sprintf (buffer, "%d", result);
    Tcl_SetErrorCode (interp, "GPGME", fnx, buffer, cp, NULL);

    return TCL_ERROR;
}


/* callback management */

static Tcl_Obj *
ctx2Obj (gpgme_ctx_t *gd) {
    char        *cp = Tcl_GetCommandName (gd -> gd_interp, gd -> gd_token);
    Tcl_Obj *resultPtr = Tcl_NewObj ();

    if (*cp != ':')
        Tcl_AppendToObj (resultPtr, "::gpgme::", -1);
    Tcl_AppendToObj (resultPtr, cp, -1);

    return resultPtr;
}


static const char *
ctx_passphrase_cb (void       *cb_value,
                   const char *desc,
                   void      **r_hd) {
    int           len;
    char         *cp;
    gpgme_ctx_t  *gd = (gpgme_ctx_t *) cb_value;
    Tcl_Interp   *interp = gd -> gd_interp;
    Tcl_Obj      *argPtr,
                 *cmdPtr;

    if (!desc)
        return NULL;

    argPtr = Tcl_NewObj ();

    Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewStringObj ("token", -1));
    Tcl_ListObjAppendElement (interp, argPtr, ctx2Obj (gd));

    if (desc != NULL) {
        Tcl_ListObjAppendElement (interp, argPtr,
                                  Tcl_NewStringObj ("description", -1));
        Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewStringObj (desc, -1));
    }

    cp = Tcl_GetByteArrayFromObj (gd -> gd_passphrase_cb, &len);
    cmdPtr = Tcl_DuplicateObj (gd -> gd_passphrase_cb);
    if ((Tcl_ListObjAppendElement (interp, cmdPtr, argPtr) != TCL_OK)
            || (Tcl_EvalObjEx (interp, cmdPtr, TCL_EVAL_DIRECT) != TCL_OK)) {
        gpgme_cancel (gd -> gd_context);
        return NULL;
    }

    if (gd -> gd_passphrase)
        Tcl_Free (gd -> gd_passphrase);
    cp = Tcl_GetByteArrayFromObj (Tcl_GetObjResult (interp), &len);
    memcpy (gd -> gd_passphrase = Tcl_Alloc (len + 1), cp, len);
    gd -> gd_passphrase[len] = '\0';

    return gd -> gd_passphrase;
}

static void
ctx_progress_cb (void       *cb_value,
                 const char *what,
                 int         type,
                 int         current,
                 int         total) {
    gpgme_ctx_t  *gd = (gpgme_ctx_t *) cb_value;
    Tcl_Interp   *interp = gd -> gd_interp;
    Tcl_Obj      *argPtr,
                 *cmdPtr;

    argPtr = Tcl_NewObj ();

    Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewStringObj ("token", -1));
    Tcl_ListObjAppendElement (interp, argPtr, ctx2Obj (gd));

    Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewStringObj ("what", -1));
    Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewStringObj (what, -1));

    Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewStringObj ("type", -1));
    Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewIntObj (type));

    Tcl_ListObjAppendElement (interp, argPtr,
                              Tcl_NewStringObj ("current", -1));
    Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewIntObj (current));

    Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewStringObj ("total", -1));
    Tcl_ListObjAppendElement (interp, argPtr, Tcl_NewIntObj (total));

    cmdPtr = Tcl_DuplicateObj (gd -> gd_progress_cb);
    if (Tcl_ListObjAppendElement (interp, cmdPtr, argPtr) != TCL_OK)
        Tcl_EvalObjEx (interp, cmdPtr, TCL_EVAL_DIRECT);
}


static void
ctx_idle_cb () {
    Tcl_HashEntry  *entryPtr;
    Tcl_HashSearch  hs,
                   *searchPtr = &hs;

    for (entryPtr = Tcl_FirstHashEntry (&ctxTable, searchPtr);
             entryPtr != NULL;
             entryPtr = Tcl_NextHashEntry (searchPtr)) {
        gpgme_ctx_t *gd = (gpgme_ctx_t *) Tcl_GetHashValue (entryPtr);
        Tcl_Interp  *interp = gd -> gd_interp;
        Tcl_Obj     *argPtr,
                    *cmdPtr;

        if (gd -> gd_idle_cb == NULL)
            continue;

        argPtr = Tcl_NewObj ();

        Tcl_ListObjAppendElement (interp, argPtr,
                                  Tcl_NewStringObj ("token", -1));
        Tcl_ListObjAppendElement (interp, argPtr, ctx2Obj (gd));

        cmdPtr = Tcl_DuplicateObj (gd -> gd_idle_cb);
        if (Tcl_ListObjAppendElement (interp, cmdPtr, argPtr) != TCL_OK)
          Tcl_EvalObjEx (interp, cmdPtr, TCL_EVAL_DIRECT);
    }
}


/* common operations */

static int
aux_proc (ClientData  data,
          Tcl_Interp *interp,
          int         objc,
          Tcl_Obj    *CONST objv[]) {
    int             argc,
                    offset,
                    result;
    char           *cp;
    oc_pair_t      *pp;
    gpgme_common_t *gd = (gpgme_common_t *) data;
    Tcl_Obj        *CONST *argv;

    Tcl_ResetResult (interp);
    argc = objc, argv = objv;
    for (argc--, argv++; argc > 0; argc -= 2, argv += 2)
        if (Tcl_StringCaseMatch (Tcl_GetString (*argv), "-operation", 0))
            break;
    if (argc <= 0) {
        Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                          " -operation", NULL);
        for (pp = gd -> gc_dispatch, cp = " "; pp -> p_key; pp++, cp = "|")
            Tcl_AppendResult (interp, cp, pp -> p_key, NULL);
        Tcl_AppendResult (interp, " ?args...?", NULL);
        return TCL_ERROR;
    }

    argc--, argv++;
    if (!*argv) {
        Tcl_AppendResult (interp, "missing argument to: \"-operation\"", NULL);
        return TCL_ERROR;
    }
    if ((result = tcl_GetIndexFromObjStruct (interp, *argv,
                                             (char **) gd -> gc_dispatch,
                                             sizeof *gd -> gc_dispatch,
                                             "operation", 0, &offset))
            != TCL_OK)
        return result;
    pp = gd -> gc_dispatch + offset;

    return ((*(pp -> p_proc)) (data, interp, objc, objv));
};


static int
aux_info (ClientData  data,
          Tcl_Interp *interp,
          int         objc,
          Tcl_Obj    *CONST objv[]) {
    oc_pair_t      *pp;
    gpgme_common_t *gd = (gpgme_common_t *) data;
    Tcl_Obj        *resultPtr;

    if (objc != 3) {
        Tcl_ResetResult (interp);
        Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                          " -operation info", NULL);
        return TCL_ERROR;
    }

    resultPtr = Tcl_NewObj ();
    for (pp = gd -> gc_dispatch; pp -> p_key; pp++)
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj (pp -> p_key, -1));

    Tcl_SetObjResult (interp, resultPtr);
    return TCL_OK;
}

/* context management */

static const char *
add_key (Tcl_Interp   *interp,
         gpgme_ctx_t  *gd,
         GpgmeKey      key,
         const char   *fpr) {
    int            result;
    Tcl_HashEntry *entryPtr;

    if ((fpr == NULL)
            && ((fpr = gpgme_key_get_string_attr (key, GPGME_ATTR_FPR, NULL,
                                                  0)) == NULL)) {
        gpgme_key_release (key);
        (void) c2t_result (interp, "gpgme_key_get_string_attr",
                           GPGME_Invalid_Value);
        return NULL;
    }

    if ((entryPtr = Tcl_FindHashEntry (&gd -> gd_keyTable, fpr)) != NULL)
        gpgme_key_release (key);
    else {
        entryPtr = Tcl_CreateHashEntry (&gd -> gd_keyTable, fpr, &result);
        Tcl_SetHashValue (entryPtr, (ClientData) key);
    }

    return fpr;
}

static GpgmeKey
find_key (Tcl_Interp  *interp,
          gpgme_ctx_t *gd,
          Tcl_Obj     *fpr) {
    Tcl_HashEntry  *entryPtr;

    if ((entryPtr = Tcl_FindHashEntry (&gd -> gd_keyTable,
                                       Tcl_GetString (fpr))) != NULL)
        return ((GpgmeKey) Tcl_GetHashValue (entryPtr));

    (void) c2t_result (interp, "find_key", GPGME_Invalid_Value);
    return NULL;
}


static Tcl_Obj *
add_trustitem (Tcl_Interp    *interp,
               gpgme_ctx_t   *gd,
               GpgmeTrustItem item) {
    int            result;
    char           buffer[BUFSIZ];
    Tcl_HashEntry *entryPtr;
    static int     ntrust = 0;

    sprintf (buffer, "trust_%d", ntrust++);
    entryPtr = Tcl_CreateHashEntry (&gd -> gd_trustTable, buffer, &result);
    Tcl_SetHashValue (entryPtr, (ClientData) item);

    return Tcl_NewStringObj (buffer, -1);
}

static GpgmeTrustItem
find_trustitem (Tcl_Interp  *interp,
                gpgme_ctx_t *gd,
                Tcl_Obj     *item) {
    Tcl_HashEntry  *entryPtr;

    if ((entryPtr = Tcl_FindHashEntry (&gd -> gd_keyTable,
                                       Tcl_GetString (item))) != NULL)
        return ((GpgmeTrustItem) Tcl_GetHashValue (entryPtr));

    (void) c2t_result (interp, "find_trustitem", GPGME_Invalid_Value);
    return NULL;
}


static Tcl_Obj *
vfy2obj (Tcl_Interp  *interp,
         gpgme_ctx_t *gd,
         int          idx) {
    unsigned long  ul;
    char          *cp;
    const char    *fpr;
    GpgmeKey       key;
    kv_pair_t     *p;
    Tcl_Obj       *attrPtr,
                  *elemPtr;

    if ((fpr = gpgme_get_sig_status (gd -> gd_context, idx, NULL, NULL))
            == NULL)
        return NULL;

    elemPtr = Tcl_NewObj ();

    for (p = ss_attrs; p -> p_key; p++) {
        ul = gpgme_get_sig_ulong_attr (gd -> gd_context, idx, p -> p_value, 0);

        switch (p -> p_value) {
            case GPGME_ATTR_VALIDITY:
                attrPtr = c2t_int (interp, validity_pairs, ul);
                break;

            case GPGME_ATTR_SIG_STATUS:
                attrPtr = c2t_int (interp, ss_pairs, ul);
                break;

            case GPGME_ATTR_SIG_SUMMARY:
                attrPtr = c2t_flags (interp, ss_flags, ul);
                break;

            default:
                if (ul) {
                    attrPtr = Tcl_NewLongObj (ul);
                    break;
                }

                if (((cp = (char *) gpgme_get_sig_string_attr
                                        (gd -> gd_context, idx, p -> p_value,
                                         0)) == NULL)
                            || (*cp == NULL))
                      continue;
                attrPtr = Tcl_NewStringObj (cp, -1);
        }

        Tcl_ListObjAppendElement (interp, elemPtr,
                                  Tcl_NewStringObj (p -> p_key, -1));
        Tcl_ListObjAppendElement (interp, elemPtr, attrPtr);
    }

    if (gpgme_get_sig_key (gd -> gd_context, idx, &key) == GPGME_No_Error) {
        Tcl_ListObjAppendElement (interp, elemPtr,
                                  Tcl_NewStringObj ("key", -1));
        Tcl_ListObjAppendElement (interp, elemPtr, Tcl_NewStringObj (fpr, -1));

        (void) add_key (interp, gd, key, fpr);
    }

    return elemPtr;
}


static int
ctx_cancel (ClientData  data,
            Tcl_Interp *interp,
            int         objc,
            Tcl_Obj    *CONST objv[]) {
    gpgme_ctx_t    *gd = (gpgme_ctx_t *) data;

    if (objc != 3) {
        Tcl_ResetResult (interp);
        Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                          " -operation cancel", NULL);
        return TCL_ERROR;
    }

    gpgme_cancel (gd -> gd_context);

    Tcl_ResetResult (interp);
    return TCL_OK;
}


#ifdef  notyet
#define WAIT_OPERATION  0
#define WAIT_BLOCKING   1

#define WAIT_MAXARGS    2
#define WAIT_OPTIONAL   WAIT_BLOCKING

static kv_pair_t wait_args[] = {
    { "-operation", WAIT_OPERATION },
    { "-blocking" , WAIT_BLOCKING  },

    { NULL        , 0              }
};

static int
ctx_wait (ClientData  data,
          Tcl_Interp *interp,
          int         objc,
          Tcl_Obj    *CONST objv[]) {
    int             blockP,
                    result;
    GpgmeCtx        ctxPtr;
    GpgmeError      status;
    gpgme_ctx_t    *gd = (gpgme_ctx_t *) data;
    Tcl_HashEntry  *entryPtr;
    Tcl_Obj        *args[WAIT_MAXARGS];

    memset (args, 0, sizeof *args * WAIT_MAXARGS);
    if (gd != NULL) {
        if ((result = crack_args (interp, objc, objv, wait_args,
                                  WAIT_OPTIONAL, args)) != TCL_OK)
            return result;
        ctxPtr = gd -> gd_context;
    } else {
        if ((result = crack_args (interp, objc, objv, wait_args + 1,
                                  WAIT_OPTIONAL - 1, args + 1)) != TCL_OK)
            return result;
        ctxPtr = NULL;
    }

    if (args[WAIT_BLOCKING] == NULL)
        blockP = 0;
    else
        if ((result = Tcl_GetBooleanFromObj (interp, args[WAIT_BLOCKING],
                                             &blockP)) != TCL_OK)
            return result;

    if ((ctxPtr = gpgme_wait (ctxPtr, &status, blockP)) == NULL)
        return c2t_result (interp, "gpgme_wait", result);

    if ((entryPtr = Tcl_FindHashEntry (&ctxTable, (char *) ctxPtr)) == NULL)
        return c2t_result (interp, "ctx_wait", GPGME_Invalid_Value);

    Tcl_SetObjResult (interp,
                      ctx2Obj ((gpgme_ctx_t *) Tcl_GetHashValue (entryPtr)));
    return TCL_OK;
}
#endif


#define GETSET_OPERATION  0
#define GETSET_PROPERTY   1
#define GETSET_VALUE      2

#define GETSET_MAXARGS    3
#define GETSET_OPTIONAL   GETSET_PROPERTY

static kv_pair_t getset_args[] = {
    { "-operation", GETSET_OPERATION },
    { "-property" , GETSET_PROPERTY  },
    { "-value"    , GETSET_VALUE  },

    { NULL        , 0              }
};

#define GET_PROTOCOL       0
#define GET_ARMOR          1
#define GET_TEXTMODE       2
#define GET_NUMBER_CERTS   3
#define GET_KEYLISTMODE    4
#define GET_PASSPHRASE_CB  5
#define GET_PROGRESS_CB    6
#define GET_IDLE_CB        7
#define GET_LAST_OP_INFO   8
#define GET_SIGNERS        9
#define SET_PROTOCOL      10
#define SET_ARMOR         11
#define SET_TEXTMODE      12
#define SET_NUMBER_CERTS  13
#define SET_KEYLISTMODE   14
#define SET_PASSPHRASE_CB 15
#define SET_PROGRESS_CB   16
#define SET_IDLE_CB       17
#define SET_SIGNERS       18

#define SET_OFFSET        SET_PROTOCOL

static kv_pair_t get_properties[] = {
    { "protocol"           , GET_PROTOCOL      },
    { "armor"              , GET_ARMOR         },
    { "textmode"           , GET_TEXTMODE      },
    { "number-of-certs"    , GET_NUMBER_CERTS  },
    { "keylistmode"        , GET_KEYLISTMODE   },
    { "passphrase-callback", GET_PASSPHRASE_CB },
    { "progress-callback"  , GET_PROGRESS_CB   },
    { "idle-callback"      , GET_IDLE_CB       },
    { "last-op-info"       , GET_LAST_OP_INFO  },
    { "signers"            , GET_SIGNERS       },

    { NULL                 , 0                 }
};

static kv_pair_t set_properties[] = {
    { "protocol"           , SET_PROTOCOL      },
    { "armor"              , SET_ARMOR         },
    { "textmode"           , SET_TEXTMODE      },
    { "number-of-certs"    , SET_NUMBER_CERTS  },
    { "keylistmode"        , SET_KEYLISTMODE   },
    { "passphrase-callback", SET_PASSPHRASE_CB },
    { "progress-callback"  , SET_PROGRESS_CB   },
    { "idle-callback"      , SET_IDLE_CB       },
    { "signers"            , SET_SIGNERS       },

    { NULL                 , 0                 }
};

static int
ctx_getset (ClientData  data,
            Tcl_Interp *interp,
            int         objc,
            Tcl_Obj    *CONST objv[]) {
    int             nelem,
                    offset,
                    onoff,
                    result;
    const char     *fpr;
    char           *cp;
    gpgme_ctx_t    *gd = (gpgme_ctx_t *) data;
    GpgmeKey        key;
    kv_pair_t      *p;
    Tcl_Obj        *elemPtr,
                   *resultPtr,
                   *args[GETSET_MAXARGS];

    memset (args, 0, sizeof *args * GETSET_MAXARGS);
    if ((result = crack_args (interp, objc, objv, getset_args,
                              GETSET_OPTIONAL, args)) != TCL_OK)
        return result;

    if (args[GETSET_PROPERTY] == NULL) {
        resultPtr = Tcl_NewObj ();
        p = (Tcl_StringCaseMatch (Tcl_GetString (args[GETSET_OPERATION]),
                                  "get", 0)) ? get_properties : set_properties;
        for (; p -> p_key; p++)
            Tcl_ListObjAppendElement (interp, resultPtr,
                                      Tcl_NewStringObj (p -> p_key, -1));

        Tcl_SetObjResult (interp, resultPtr);
        return TCL_OK;
    }

    if (Tcl_StringCaseMatch (Tcl_GetString (args[GETSET_OPERATION]), "get", 0))
        p = get_properties;
    else
        if (args[GETSET_VALUE] == NULL) {
            Tcl_ResetResult (interp);
            Tcl_AppendResult (interp, getset_args[GETSET_VALUE].p_key,
                              " switch must be provided", NULL);
            return TCL_ERROR;
        }
        else
            p = set_properties;

    if ((result = tcl_GetIndexFromObjStruct (interp, args[GETSET_PROPERTY],
                                             (char **) p,
                                             sizeof *p, "property",
                                             0, &offset)) != TCL_OK)
        return result;
    offset = p[offset].p_value;

    switch (offset) {
        case SET_PASSPHRASE_CB:
        case SET_PROGRESS_CB:
        case SET_IDLE_CB:
        case SET_SIGNERS:
            if ((result = Tcl_ListObjLength (interp, args[GETSET_VALUE],
                                             &nelem)) != TCL_OK)
                return result;
            break;
    }

    resultPtr = NULL;
    switch (offset) {
        case GET_PROTOCOL:
            resultPtr = c2t_int (interp, protocol_pairs,
                                 gpgme_get_protocol (gd -> gd_context));
            break;

        case SET_PROTOCOL:
            if ((result = tcl_GetIndexFromObjStruct (interp,
                                                     args[GETSET_VALUE],
                                                     (char **) protocol_pairs,
                                                     sizeof *protocol_pairs,
                                                     "protocol", 0, &offset))
                    != TCL_OK)
                return result;
            gpgme_set_protocol (gd -> gd_context,
                                protocol_pairs[offset].p_value);
            break;

        case GET_ARMOR:
            resultPtr = Tcl_NewBooleanObj (gpgme_get_armor (gd -> gd_context));
            break;

        case SET_ARMOR:
            if ((result = Tcl_GetBooleanFromObj (interp, args[GETSET_VALUE],
                                                 &onoff)) != TCL_OK)
                return result;
            gpgme_set_armor (gd -> gd_context, onoff);
            break;

        case GET_TEXTMODE:
            resultPtr = Tcl_NewBooleanObj
                                (gpgme_get_textmode (gd -> gd_context));
            break;

        case SET_TEXTMODE:
            if ((result = Tcl_GetBooleanFromObj (interp, args[GETSET_VALUE],
                                                 &onoff)) != TCL_OK)
                return result;
            gpgme_set_textmode (gd -> gd_context, onoff);
            break;

        case GET_NUMBER_CERTS:
            resultPtr = Tcl_NewIntObj
                            (gpgme_get_include_certs (gd -> gd_context));
            break;

        case SET_NUMBER_CERTS:
            if ((result = Tcl_GetIntFromObj (interp, args[GETSET_VALUE],
                                             &nelem)) != TCL_OK)
                return result;
            gpgme_set_include_certs (gd -> gd_context, nelem);
            break;

        case GET_KEYLISTMODE:
            resultPtr = c2t_flags (interp, keylist_flags,
                                   gpgme_get_keylist_mode (gd -> gd_context));
            break;

        case SET_KEYLISTMODE:
            if ((result = t2c_flags (interp, args[GETSET_VALUE], keylist_flags,
                                     &onoff)) != TCL_OK)
                return result;
            gpgme_set_keylist_mode (gd -> gd_context, onoff);
            break;

        case GET_PASSPHRASE_CB:
            if (gd -> gd_passphrase_cb == NULL)
                resultPtr = Tcl_NewObj ();
            else
                Tcl_IncrRefCount (resultPtr = gd -> gd_passphrase_cb);      
            break;

        case SET_PASSPHRASE_CB:
            if (gd -> gd_passphrase_cb != NULL)
                Tcl_DecrRefCount (gd -> gd_passphrase_cb);
            if (nelem > 0) {
                Tcl_IncrRefCount (gd -> gd_passphrase_cb = args[GETSET_VALUE]);
                gpgme_set_passphrase_cb (gd -> gd_context, ctx_passphrase_cb,
                                         gd);
            } else {
                gd -> gd_passphrase_cb = NULL;
                gpgme_set_passphrase_cb (gd -> gd_context, NULL, NULL);
            }
            break;

        case GET_PROGRESS_CB:
            if (gd -> gd_progress_cb == NULL)
                resultPtr = Tcl_NewObj ();
            else
                Tcl_IncrRefCount (resultPtr = gd -> gd_progress_cb);        
            break;

        case SET_PROGRESS_CB:
            if (gd -> gd_progress_cb != NULL)
                Tcl_DecrRefCount (gd -> gd_progress_cb);
            if (nelem > 0) {
                Tcl_IncrRefCount (gd -> gd_progress_cb = args[GETSET_VALUE]);
                gpgme_set_progress_cb (gd -> gd_context, ctx_progress_cb,
                                       gd);
            } else {
                gd -> gd_progress_cb = NULL;
                gpgme_set_progress_cb (gd -> gd_context, NULL, NULL);
            }
            break;

        case GET_IDLE_CB:
            if (gd -> gd_idle_cb == NULL)
                resultPtr = Tcl_NewObj ();
            else
                Tcl_IncrRefCount (resultPtr = gd -> gd_idle_cb);        
            break;

        case SET_IDLE_CB:
            if (gd -> gd_idle_cb != NULL)
                Tcl_DecrRefCount (gd -> gd_idle_cb);
            if (nelem > 0)
                Tcl_IncrRefCount (gd -> gd_idle_cb = args[GETSET_VALUE]);
            else
                gd -> gd_idle_cb = NULL;
            break;

        case GET_LAST_OP_INFO:
            resultPtr =
                Tcl_NewStringObj (cp = gpgme_get_op_info (gd -> gd_context,
                                                          0), -1);
            free (cp);
            break;

        case GET_SIGNERS:
            resultPtr = Tcl_NewObj ();
            for (offset = 0;
                     (key = gpgme_signers_enum (gd ->gd_context, offset))
                            != NULL;
                     offset++)
                if ((fpr = add_key (interp, gd, key, NULL)) == NULL)
                    return TCL_ERROR;
                else
                    Tcl_ListObjAppendElement (interp, resultPtr,
                                              Tcl_NewStringObj (fpr, -1));
            break;

        case SET_SIGNERS:
            for (offset = 0; offset < nelem; offset++)
                if ((result = Tcl_ListObjIndex (interp, args[GETSET_VALUE],
                                                offset, &elemPtr)) != TCL_OK)
                    return result;
                else
                    if (find_key (interp, gd, elemPtr) == NULL)
                        return TCL_ERROR;

            gpgme_signers_clear (gd -> gd_context);
            for (offset = 0; offset < nelem; offset++)
                if (((result = Tcl_ListObjIndex (interp, args[GETSET_VALUE],
                                                 offset, &elemPtr)) == TCL_OK)
                        && ((key = find_key (interp, gd, elemPtr)) != NULL)
                        && ((result = gpgme_signers_add (gd -> gd_context,
                                                         key))
                                != GPGME_No_Error))
                    return c2t_result (interp, "gpgme_signers_add", result);
            break;
    }

    if (resultPtr != NULL)
        Tcl_SetObjResult (interp, resultPtr);
    else
        Tcl_ResetResult (interp);
    return TCL_OK;
}


#define ENCRYPT_OPERATION  0
#define ENCRYPT_INPUT      1
#define ENCRYPT_RECIPIENTS 2
#define ENCRYPT_SIGN       3

#define ENCRYPT_MAXARGS    4
#define ENCRYPT_OPTIONAL   ENCRYPT_SIGN

static kv_pair_t encrypt_args[] = {
    { "-operation" , ENCRYPT_OPERATION  },
    { "-input"     , ENCRYPT_INPUT      },
    { "-recipients", ENCRYPT_RECIPIENTS },
    { "-sign"      , ENCRYPT_SIGN       },

    { NULL         , 0                  }
};

static int
ctx_encrypt (ClientData  data,
            Tcl_Interp *interp,
            int         objc,
            Tcl_Obj    *CONST objv[]) {
    int          cc,
                 len,
                 result,
                 signP;
    char        *cp,
                 buffer[BUFSIZ];
    GpgmeData    in,
                 out;
    gpgme_ctx_t   *gd = (gpgme_ctx_t *) data;
    gpgme_recip_t *rd;
    Tcl_HashEntry *entryPtr;
    Tcl_Obj       *resultPtr,
                  *args[ENCRYPT_MAXARGS];

    memset (args, 0, sizeof *args * ENCRYPT_MAXARGS);
    if ((result = crack_args (interp, objc, objv, encrypt_args,
                              ENCRYPT_OPTIONAL, args)) != TCL_OK)
        return result;

    cp = Tcl_GetString (args[ENCRYPT_RECIPIENTS]);
    if ((entryPtr = Tcl_FindHashEntry (&recipXTable, cp)) == NULL) {
        Tcl_ResetResult (interp);
        Tcl_AppendResult (interp, "invalid recipients token: \"", cp, "\"",
                          NULL);
        return TCL_ERROR;
    }
    rd = (gpgme_recip_t *) Tcl_GetHashValue (entryPtr);

    if (args[ENCRYPT_SIGN] == NULL)
        signP = 0;
    else
        if ((result = Tcl_GetBooleanFromObj (interp, args[ENCRYPT_SIGN],
                                             &signP)) != TCL_OK)
            return result;

    cp = Tcl_GetByteArrayFromObj (args[ENCRYPT_INPUT], &len);
    if ((result = gpgme_data_new_from_mem (&in, cp, len, 0)) != GPGME_No_Error)
        return c2t_result (interp, "gpgme_data_new_from_mem", result);

    if ((result = gpgme_data_new (&out)) != GPGME_No_Error) {
        gpgme_data_release (in);
        return c2t_result (interp, "gpgme_data_new", result);
    }

    if (signP)
      result = gpgme_op_encrypt_sign (gd -> gd_context, rd -> gd_recipients,
                                      in, out);
    else
      result = gpgme_op_encrypt (gd -> gd_context, rd -> gd_recipients, in,
                                 out);

    gpgme_data_release (in);

    if (result != GPGME_No_Error) {
        gpgme_data_release (out);
        return c2t_result (interp, signP ? "gpgme_op_encrypt_sign"
                                         : "gpgme_op_encrypt", result);
    }

    if ((result = gpgme_data_rewind (out)) != GPGME_No_Error) {
        gpgme_data_release (out);
        return c2t_result (interp, "gpgme_data_rewind", result);
    }

    Tcl_SetObjResult (interp, resultPtr = Tcl_NewByteArrayObj (NULL, cc = 0));
    while ((result = gpgme_data_read (out, buffer, sizeof buffer, &len))
                == GPGME_No_Error) {
        unsigned char *up = Tcl_SetByteArrayLength (resultPtr, cc + len);

        memcpy (up + cc, buffer, len);
        cc += len;
    }

    gpgme_data_release (out);

    if (result != GPGME_EOF)
        return c2t_result (interp, "gpgme_data_read", result);

    return TCL_OK;
}


#define DECRYPT_OPERATION   0
#define DECRYPT_INPUT       1
#define DECRYPT_CHECKSTATUS 2

#define DECRYPT_MAXARGS    3
#define DECRYPT_OPTIONAL   DECRYPT_CHECKSTATUS

static kv_pair_t decrypt_args[] = {
    { "-operation"  , DECRYPT_OPERATION   },
    { "-input"      , DECRYPT_INPUT       },
    { "-checkstatus", DECRYPT_CHECKSTATUS },

    { NULL          , 0                   }
};

static int
ctx_decrypt (ClientData  data,
            Tcl_Interp *interp,
            int         objc,
            Tcl_Obj    *CONST objv[]) {
    int          cc,
                 checkP,
                 len,
                 result;
    char        *cp,
                 buffer[BUFSIZ];
    GpgmeData    in,
                 out;
    GpgmeSigStat status;
    gpgme_ctx_t   *gd = (gpgme_ctx_t *) data;
    Tcl_Obj       *resultPtr,
                  *elemPtr,
                  *args[DECRYPT_MAXARGS];

    memset (args, 0, sizeof *args * DECRYPT_MAXARGS);
    if ((result = crack_args (interp, objc, objv, decrypt_args,
                              DECRYPT_OPTIONAL, args)) != TCL_OK)
        return result;

    if (args[DECRYPT_CHECKSTATUS] == NULL)
        checkP = 0;
    else
        if ((result = Tcl_GetBooleanFromObj (interp, args[DECRYPT_CHECKSTATUS],
                                             &checkP)) != TCL_OK)
            return result;

    cp = Tcl_GetByteArrayFromObj (args[DECRYPT_INPUT], &len);
    if ((result = gpgme_data_new_from_mem (&in, cp, len, 0)) != GPGME_No_Error)
        return c2t_result (interp, "gpgme_data_new_from_mem", result);

    if ((result = gpgme_data_new (&out)) != GPGME_No_Error) {
        gpgme_data_release (in);
        return c2t_result (interp, "gpgme_data_new", result);
    }

    if (checkP)
        result = gpgme_op_decrypt_verify (gd -> gd_context, in, out, &status);
    else
        result = gpgme_op_decrypt (gd -> gd_context, in, out);

    gpgme_data_release (in);

    if (result != GPGME_No_Error) {
        gpgme_data_release (out);
        return c2t_result (interp, checkP ? "gpgme_op_decrypt_verify"
                                          : "gpgme_op_decrypt", result);
    }

    if ((result = gpgme_data_rewind (out)) != GPGME_No_Error) {
        gpgme_data_release (out);
        return c2t_result (interp, "gpgme_data_rewind", result);
    }

    Tcl_SetObjResult (interp, resultPtr = Tcl_NewObj ());

    Tcl_ListObjAppendElement (interp, resultPtr,
                              Tcl_NewStringObj ("plaintext", -1));
    Tcl_ListObjAppendElement (interp, resultPtr,
                              elemPtr = Tcl_NewByteArrayObj (NULL, cc = 0));
    while ((result = gpgme_data_read (out, buffer, sizeof buffer, &len))
                == GPGME_No_Error) {
        unsigned char *up = Tcl_SetByteArrayLength (elemPtr, cc + len);

        memcpy (up + cc, buffer, len);
        cc += len;
    }

    gpgme_data_release (out);

    if (result != GPGME_EOF)
        return c2t_result (interp, "gpgme_data_read", result);

    if (checkP) {
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj ("status", -1));
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  c2t_int (interp, ss_pairs, status));
    }

    return TCL_OK;
}


#define SIGN_OPERATION 0
#define SIGN_INPUT     1
#define SIGN_MODE      2

#define SIGN_MAXARGS   3
#define SIGN_OPTIONAL  SIGN_MODE

static kv_pair_t sign_args[] = {
    { "-operation", SIGN_OPERATION },
    { "-input"    , SIGN_INPUT     },
    { "-mode"     , SIGN_MODE      },

    { NULL        , 0              }
};

static int
ctx_sign (ClientData  data,
          Tcl_Interp *interp,
          int         objc,
          Tcl_Obj    *CONST objv[]) {
    int          cc,
                 len,
                 mode,
                 result;
    char        *cp,
                 buffer[BUFSIZ];
    GpgmeData    in,
                 out;
    gpgme_ctx_t *gd = (gpgme_ctx_t *) data;
    Tcl_Obj     *resultPtr,
                *args[SIGN_MAXARGS];

    memset (args, 0, sizeof *args * SIGN_MAXARGS);
    if ((result = crack_args (interp, objc, objv, sign_args,
                              SIGN_OPTIONAL, args)) != TCL_OK)
        return result;

    if (args[SIGN_MODE] == NULL)
        mode = GPGME_SIG_MODE_NORMAL;
    else {
        if ((result = tcl_GetIndexFromObjStruct (interp, args[SIGN_MODE],
                                                 (char **) sm_pairs,
                                                 sizeof *sm_pairs, "mode",
                                                 0, &mode)) != TCL_OK)
            return result;

        mode = sm_pairs[mode].p_value;
    }

    cp = Tcl_GetByteArrayFromObj (args[SIGN_INPUT], &len);
    if ((result = gpgme_data_new_from_mem (&in, cp, len, 0)) != GPGME_No_Error)
        return c2t_result (interp, "gpgme_data_new_from_mem", result);

    if ((result = gpgme_data_new (&out)) != GPGME_No_Error) {
        gpgme_data_release (in);
        return c2t_result (interp, "gpgme_data_new", result);
    }

    result = gpgme_op_sign (gd -> gd_context, in, out, mode);

    gpgme_data_release (in);

    if (result != GPGME_No_Error) {
        gpgme_data_release (out);
        return c2t_result (interp, "gpgme_op_sign", result);
    }

    if ((result = gpgme_data_rewind (out)) != GPGME_No_Error) {
        gpgme_data_release (out);
        return c2t_result (interp, "gpgme_data_rewind", result);
    }

    Tcl_SetObjResult (interp, resultPtr = Tcl_NewByteArrayObj (NULL, cc = 0));
    while ((result = gpgme_data_read (out, buffer, sizeof buffer, &len))
                == GPGME_No_Error) {
        unsigned char *up = Tcl_SetByteArrayLength (resultPtr, cc + len);

        memcpy (up + cc, buffer, len);
        cc += len;
    }

    gpgme_data_release (out);

    if (result != GPGME_EOF)
        return c2t_result (interp, "gpgme_data_read", result);

    return TCL_OK;
}


#define VERIFY_OPERATION 0
#define VERIFY_SIGNATURE 1
#define VERIFY_INPUT     2

#define VERIFY_MAXARGS    3
#define VERIFY_OPTIONAL   VERIFY_INPUT

static kv_pair_t verify_args[] = {
    { "-operation", VERIFY_OPERATION },
    { "-signature", VERIFY_SIGNATURE },
    { "-input"    , VERIFY_INPUT     },

    { NULL        , 0                }
};

static int
ctx_verify (ClientData  data,
            Tcl_Interp *interp,
            int         objc,
            Tcl_Obj    *CONST objv[]) {
    int          cc,
                 len,
                 offset,
                 result;
    char        *cp,
                 buffer[BUFSIZ];
    GpgmeData    sig,
                 text;
    GpgmeSigStat status;
    gpgme_ctx_t *gd = (gpgme_ctx_t *) data;
    Tcl_Obj     *attrPtr,
                *elemPtr,
                *resultPtr,
                *args[VERIFY_MAXARGS];

    memset (args, 0, sizeof *args * VERIFY_MAXARGS);
    if ((result = crack_args (interp, objc, objv, verify_args,
                              VERIFY_OPTIONAL, args)) != TCL_OK)
        return result;

    cp = Tcl_GetByteArrayFromObj (args[VERIFY_SIGNATURE], &len);
    if ((result = gpgme_data_new_from_mem (&sig, cp, len, 0))
            != GPGME_No_Error)
        return c2t_result (interp, "gpgme_data_new_from_mem", result);

    if (args[VERIFY_INPUT] != NULL) {
        cp = Tcl_GetByteArrayFromObj (args[VERIFY_INPUT], &len);
        result = gpgme_data_new_from_mem (&text, cp, len, 0);
    } else
        result = gpgme_data_new (&text);
    if (result != GPGME_No_Error) {
        gpgme_data_release (sig);
        return c2t_result (interp, args[VERIFY_INPUT] != NULL
                                       ? "gpgme_data_new_from_mem"
                                       : "gpgme_data_new", result);
    }

    result = gpgme_op_verify (gd -> gd_context, sig, text, &status);

    gpgme_data_release (sig);

    if (result != GPGME_No_Error) {
        gpgme_data_release (text);
        return c2t_result (interp, "gpgme_op_verify", result);
    }
    
    Tcl_SetObjResult (interp, resultPtr = Tcl_NewObj ());

    if (args[VERIFY_INPUT] == NULL) {
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj ("plaintext", -1));
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  elemPtr = Tcl_NewByteArrayObj (NULL,
                                                                 cc = 0));
        while ((result = gpgme_data_read (text, buffer, sizeof buffer, &len))
                   == GPGME_No_Error) {
            unsigned char *up = Tcl_SetByteArrayLength (elemPtr, cc + len);

            memcpy (up + cc, buffer, len);
            cc += len;
        }
    }
    else
        result = GPGME_EOF;

    gpgme_data_release (text);

    if (result != GPGME_EOF)
        return c2t_result (interp, "gpgme_data_read", result);

    Tcl_ListObjAppendElement (interp, resultPtr,
                              Tcl_NewStringObj ("status", -1));
    Tcl_ListObjAppendElement (interp, resultPtr,
                              c2t_int (interp, ss_pairs, status));

    if ((cp = gpgme_get_notation (gd -> gd_context)) != NULL) {
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj ("notation", -1));
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj (cp, -1));
    }

    Tcl_ListObjAppendElement (interp, resultPtr,
                              Tcl_NewStringObj ("signatures", -1));
    Tcl_ListObjAppendElement (interp, resultPtr, elemPtr = Tcl_NewObj ());
    for (offset = 0;
             (attrPtr = vfy2obj (interp, gd, offset)) != NULL;
             offset++)
        Tcl_ListObjAppendElement (interp, elemPtr, attrPtr);


    return TCL_OK;
}


#define STARTKEY_OPERATION  0
#define STARTKEY_PATTERNS   1
#define STARTKEY_SECRETONLY 2

#define STARTKEY_MAXARGS    3
#define STARTKEY_OPTIONAL   STARTKEY_PATTERNS

static kv_pair_t startkey_args[] = {
    { "-operation" , STARTKEY_OPERATION  },
    { "-patterns"  , STARTKEY_PATTERNS   },
    { "-secretonly", STARTKEY_SECRETONLY },

    { NULL         , 0                   }
};

static int
ctx_start_key (ClientData  data,
               Tcl_Interp *interp,
               int         objc,
               Tcl_Obj    *CONST objv[]) {
    int          i,
                 nelem,
                 result,
                 secretP; 
    char       **ap,
               **patterns;
    gpgme_ctx_t *gd = (gpgme_ctx_t *) data;
    Tcl_Obj     *elemPtr,
                *args[STARTKEY_MAXARGS];

    memset (args, 0, sizeof *args * STARTKEY_MAXARGS);
    if ((result = crack_args (interp, objc, objv, startkey_args,
                              STARTKEY_OPTIONAL, args)) != TCL_OK)
        return result;

    if (gd -> gd_flags & GD_KEYLIST) {
        Tcl_ResetResult (interp);
        Tcl_SetResult (interp,
                       "already doing a key listing, end that one first",
                       TCL_STATIC);
        return TCL_ERROR;
    }

    if (args[STARTKEY_PATTERNS] != NULL) {
        if ((result = Tcl_ListObjLength (interp, args[STARTKEY_PATTERNS],
                                         &nelem)) != TCL_OK)
            return result;
    }
    else
        nelem = 0;

    if (args[STARTKEY_SECRETONLY] == NULL)
        secretP = 0;
    else
        if ((result = Tcl_GetBooleanFromObj (interp, args[STARTKEY_SECRETONLY],
                                             &secretP)) != TCL_OK)
            return result;

    Tcl_ResetResult (interp);
    if (nelem == 0) {
        if ((result = gpgme_op_keylist_start (gd -> gd_context, NULL, secretP))
                != GPGME_No_Error)
            return c2t_result (interp, "gpgme_op_keylist_start", result);
        gd -> gd_flags |= GD_KEYLIST;

        return TCL_OK;
    }
    
    patterns = ap = (char **) Tcl_Alloc (sizeof *patterns * (nelem + 1));
    for (i = 0; i < nelem; i++) {
        if ((result = Tcl_ListObjIndex (interp, args[STARTKEY_PATTERNS], i,
                                        &elemPtr)) != TCL_OK) {
            Tcl_Free ((char *) patterns);
            return result;
        }

        *ap++ = Tcl_GetString (elemPtr);
    }
    *ap = NULL;

    if ((result = gpgme_op_keylist_ext_start (gd -> gd_context,
                                              (const char **) patterns,
                                              secretP, 0)) == GPGME_No_Error)
        gd -> gd_flags |= GD_KEYLIST;

    Tcl_Free ((char *) patterns);

    return c2t_result (interp, "gpgme_op_keylist_ext_start", result);
}


static int
ctx_next_key (ClientData  data,
              Tcl_Interp *interp,
              int         objc,
              Tcl_Obj    *CONST objv[]) {
    int            result; 
    const char    *fpr;
    gpgme_ctx_t   *gd = (gpgme_ctx_t *) data;
    GpgmeKey       key;

    if (objc != 3) {
        Tcl_ResetResult (interp);
        Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                          " -operation next-key", NULL);
        return TCL_ERROR;
    }

    if (!(gd -> gd_flags & GD_KEYLIST)) {
        Tcl_ResetResult (interp);
        Tcl_SetResult (interp, "not doing a key listing", TCL_STATIC);
        return TCL_ERROR;
    }

    switch (result = gpgme_op_keylist_next (gd -> gd_context, &key)) {
        case GPGME_EOF:
            gd -> gd_flags |= GD_KEYEOF;
            Tcl_ResetResult (interp);
            return TCL_OK;

        case GPGME_No_Error:
            if ((fpr = add_key (interp, gd, key, NULL)) == NULL)
                return TCL_ERROR;

            Tcl_SetObjResult (interp, Tcl_NewStringObj (fpr, -1));
            return TCL_OK;

        default:
            return c2t_result (interp, "gpgme_op_keylist_next", result);
    }
}


static int
ctx_done_key (ClientData  data,
              Tcl_Interp *interp,
              int         objc,
              Tcl_Obj    *CONST objv[]) {
    gpgme_ctx_t *gd = (gpgme_ctx_t *) data;

    if (objc != 3) {
        Tcl_ResetResult (interp);
        Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                          " -operation done-key", NULL);
        return TCL_ERROR;
    }

    if (!(gd -> gd_flags & GD_KEYLIST)) {
        Tcl_ResetResult (interp);
        Tcl_SetResult (interp, "not doing a key listing", TCL_STATIC);
        return TCL_ERROR;
    }

    gd -> gd_flags &= ~GD_KEYLIST;
    if (!(gd -> gd_flags & GD_KEYEOF))
        return c2t_result (interp, "gpgme_op_keylist_end",
                           gpgme_op_keylist_end (gd -> gd_context));

    gd -> gd_flags &= ~GD_KEYEOF;

    Tcl_ResetResult (interp);
    return TCL_OK;
}


#define INFOKEY_OPERATION 0
#define INFOKEY_KEY       1

#define INFOKEY_MAXARGS    2
#define INFOKEY_OPTIONAL   INFOKEY_MAXARGS

static kv_pair_t infokey_args[] = {
    { "-operation" , INFOKEY_OPERATION },
    { "-key"       , INFOKEY_KEY       },

    { NULL         , 0                 }
};

static Tcl_Obj *
ctx_info_key_aux (Tcl_Interp *interp,
             GpgmeKey    key,
             int         idx);

static int
ctx_info_key (ClientData  data,
              Tcl_Interp *interp,
              int         objc,
              Tcl_Obj    *CONST objv[]) {
    int            offset,
                   result;
    char          *cp;
    gpgme_ctx_t   *gd = (gpgme_ctx_t *) data;
    GpgmeKey       key;
    Tcl_Obj       *attrPtr,
                  *elemPtr,
                  *resultPtr,
                  *args[INFOKEY_MAXARGS];

    memset (args, 0, sizeof *args * INFOKEY_MAXARGS);
    if ((result = crack_args (interp, objc, objv, infokey_args,
                              INFOKEY_OPTIONAL, args)) != TCL_OK)
        return result;

    if ((key = find_key (interp, gd, args[INFOKEY_KEY])) == NULL)
        return TCL_ERROR;

    Tcl_SetObjResult (interp, resultPtr = Tcl_NewObj ());

    if ((cp = gpgme_key_get_as_xml (key)) != NULL) {
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj ("xml", -1));
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj (cp, -1));
        free (cp);
    }

    Tcl_ListObjAppendElement (interp, resultPtr,
                              Tcl_NewStringObj ("subkeys", -1));
    Tcl_ListObjAppendElement (interp, resultPtr, elemPtr = Tcl_NewObj ());

    for (offset = 0;
            (attrPtr = ctx_info_key_aux (interp, key, offset)) != NULL;
            offset ++)
        Tcl_ListObjAppendElement (interp, elemPtr, attrPtr);

    return TCL_OK;
}

static Tcl_Obj *
ctx_info_key_aux (Tcl_Interp *interp,
             GpgmeKey    key,
             int         idx) {
    int      hitP;
    unsigned long  ul;
    char          *cp;
    kv_pair_t     *p;
    Tcl_Obj       *attrPtr,
                  *elemPtr;

    elemPtr = NULL;
    hitP = 0;

    for (p = attr_pairs; p -> p_key; p++) {
        attrPtr = NULL;

        ul = gpgme_key_get_ulong_attr (key, p -> p_value, NULL, idx);

        switch (p -> p_value) {
            case GPGME_ATTR_OTRUST:
                if ((ul == GPGME_VALIDITY_UNKNOWN)
                        || (ul == GPGME_VALIDITY_UNDEFINED))
                    continue;
                /* else fall... */
            case GPGME_ATTR_VALIDITY:
                attrPtr = c2t_int (interp, validity_pairs, ul);
                break;

            case GPGME_ATTR_KEY_CAPS:
                attrPtr = c2t_flags (interp, capability_flags, ul);
                break;

            default:
                if (ul) {
                    attrPtr = Tcl_NewLongObj (ul);
                    break;
                }
                /* else fall... */

            case GPGME_ATTR_ALGO:
                if (((cp = (char *) gpgme_key_get_string_attr (key,
                                                               p -> p_value,
                                                               NULL, idx))
                                == NULL)
                            || (*cp == NULL))
                    continue;
                if (p -> p_value == GPGME_ATTR_KEYID)
                    hitP++;
                attrPtr = Tcl_NewStringObj (cp, -1);
                break;
        }

        if (elemPtr == NULL)
            elemPtr = Tcl_NewObj ();

        Tcl_ListObjAppendElement (interp, elemPtr,
                                  Tcl_NewStringObj (p -> p_key, -1));
        Tcl_ListObjAppendElement (interp, elemPtr, attrPtr);
    }

    if (hitP || (elemPtr == NULL))
        return elemPtr;

    TclFreeObj (elemPtr);
    return NULL;
}


#define STARTRUST_OPERATION 0
#define STARTRUST_PATTERN   1
#define STARTRUST_MAXLEVEL  2

#define STARTRUST_MAXARGS    3
#define STARTRUST_OPTIONAL   STARTRUST_MAXLEVEL

static kv_pair_t startrust_args[] = {
    { "-operation", STARTRUST_OPERATION },
    { "-pattern"  , STARTRUST_PATTERN   },
    { "-maxlevel" , STARTRUST_MAXLEVEL  },

    { NULL         , 0                   }
};

static int
ctx_start_trustitem (ClientData  data,
                     Tcl_Interp *interp,
                     int         objc,
                     Tcl_Obj    *CONST objv[]) {
    int          maxlevel,
                 result;
    gpgme_ctx_t *gd = (gpgme_ctx_t *) data;
    Tcl_Obj     *args[STARTRUST_MAXARGS];

    memset (args, 0, sizeof *args * STARTRUST_MAXARGS);
    if ((result = crack_args (interp, objc, objv, startrust_args,
                              STARTRUST_OPTIONAL, args)) != TCL_OK)
        return result;

    if (gd -> gd_flags & GD_TRUSTLIST) {
        Tcl_ResetResult (interp);
        Tcl_SetResult (interp,
                       "already doing a trustitem listing, end that one first",
                       TCL_STATIC);
        return TCL_ERROR;
    }

    if (args[STARTRUST_MAXLEVEL] == NULL)
        maxlevel = 0;
    else
        if ((result = Tcl_GetIntFromObj (interp, args[STARTRUST_MAXLEVEL],
                                             &maxlevel)) != TCL_OK)
            return result;

    if ((result = gpgme_op_trustlist_start
                        (gd -> gd_context,
                         Tcl_GetString (args[STARTRUST_PATTERN]), maxlevel))
                != GPGME_No_Error)
        return c2t_result (interp, "gpgme_op_trustlist_start", result);

    gd -> gd_flags |= GD_TRUSTLIST;

    Tcl_ResetResult (interp);
    return TCL_OK;
}


static int
ctx_next_trustitem (ClientData  data,
                    Tcl_Interp *interp,
                    int         objc,
                    Tcl_Obj    *CONST objv[]) {
    int            result; 
    gpgme_ctx_t   *gd = (gpgme_ctx_t *) data;
    GpgmeTrustItem item;

    if (objc != 3) {
        Tcl_ResetResult (interp);
        Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                          " -operation next-trustitem", NULL);
        return TCL_ERROR;
    }

    if (!(gd -> gd_flags & GD_TRUSTLIST)) {
        Tcl_ResetResult (interp);
        Tcl_SetResult (interp, "not doing a trustitem listing", TCL_STATIC);
        return TCL_ERROR;
    }

    switch (result = gpgme_op_trustlist_next (gd -> gd_context, &item)) {
        case GPGME_EOF:
            gd -> gd_flags |= GD_TRUSTEOF;
            Tcl_ResetResult (interp);
            return TCL_OK;
            
        case GPGME_No_Error:
            Tcl_SetObjResult (interp, add_trustitem (interp, gd, item));
            return TCL_OK;

        default:
            return c2t_result (interp, "gpgme_op_trustlist_next", result);
    }
}


static int
ctx_done_trustitem (ClientData  data,
                    Tcl_Interp *interp,
                    int         objc,
                    Tcl_Obj    *CONST objv[]) {
    gpgme_ctx_t *gd = (gpgme_ctx_t *) data;

    if (objc != 3) {
        Tcl_ResetResult (interp);
        Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                          " -operation done-trustitem", NULL);
        return TCL_ERROR;
    }

    if (!(gd -> gd_flags & GD_TRUSTLIST)) {
        Tcl_ResetResult (interp);
        Tcl_SetResult (interp, "not doing a trustitem listing", TCL_STATIC);
        return TCL_ERROR;
    }

    gd -> gd_flags &= ~GD_TRUSTLIST;
    if (!(gd -> gd_flags & GD_TRUSTEOF))
        return c2t_result (interp, "gpgme_op_trustlist_end",
                           gpgme_op_trustlist_end (gd -> gd_context));

    gd -> gd_flags &= ~GD_TRUSTEOF;

    Tcl_ResetResult (interp);
    return TCL_OK;
}


#define INFOTRUST_OPERATION 0
#define INFOTRUST_ITEM      1

#define INFOTRUST_MAXARGS    2
#define INFOTRUST_OPTIONAL   INFOTRUST_MAXARGS

static kv_pair_t infotrust_args[] = {
    { "-operation" , INFOTRUST_OPERATION },
    { "-item"      , INFOTRUST_ITEM      },

    { NULL         , 0                   }
};

static int
ctx_info_trustitem (ClientData  data,
                    Tcl_Interp *interp,
                    int         objc,
                    Tcl_Obj    *CONST objv[]) {
    int            result;
    char          *cp;
    gpgme_ctx_t   *gd = (gpgme_ctx_t *) data;
    GpgmeTrustItem item;
    kv_pair_t     *p;
    Tcl_Obj       *attrPtr,
                  *resultPtr,
                  *args[INFOTRUST_MAXARGS];

    memset (args, 0, sizeof *args * INFOTRUST_MAXARGS);
    if ((result = crack_args (interp, objc, objv, infotrust_args,
                              INFOTRUST_OPTIONAL, args)) != TCL_OK)
        return result;

    if ((item = find_trustitem (interp, gd, args[INFOTRUST_ITEM])) == NULL)
        return TCL_ERROR;

    Tcl_SetObjResult (interp, resultPtr = Tcl_NewObj ());

    for (p = attr_pairs; p -> p_key; p++) {
        result = gpgme_trust_item_get_int_attr (item, p -> p_value, NULL, 0);

        switch (p -> p_value) {
            case GPGME_ATTR_TYPE:
                attrPtr = c2t_int (interp, type_pairs, result);
                break;

            case GPGME_ATTR_OTRUST:
                if ((result == GPGME_VALIDITY_UNKNOWN)
                        || (result == GPGME_VALIDITY_UNDEFINED))
                    continue;
                /* else fall... */
            case GPGME_ATTR_VALIDITY:
                attrPtr = c2t_int (interp, validity_pairs, result);
                break;

            default:
                if ((result != 0) || (p -> p_value == GPGME_ATTR_LEVEL)) {
                    attrPtr = Tcl_NewIntObj (result);
                    break;
                }
                if (((cp = (char *) gpgme_trust_item_get_string_attr
                                        (item, p -> p_value, NULL, 0)) == NULL)
                            || (*cp == NULL))
                    continue;
                attrPtr = Tcl_NewStringObj (cp, -1);
                break;
        }

        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj (p -> p_key, -1));
        Tcl_ListObjAppendElement (interp, resultPtr, attrPtr);
    }

    return TCL_OK;
}

/*  */

static void
ctx_free (ClientData data) {
    gpgme_ctx_t   *gd = (gpgme_ctx_t *) data;
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch hs,
                  *searchPtr = &hs;

    if ((entryPtr = Tcl_FindHashEntry (&ctxTable, (char *) gd -> gd_context))
            != NULL)
        Tcl_DeleteHashEntry (entryPtr);

    if (gd -> gd_passphrase_cb)
        Tcl_DecrRefCount (gd -> gd_passphrase_cb);
    if (gd -> gd_passphrase)
        Tcl_Free (gd -> gd_passphrase);
    if (gd -> gd_progress_cb)
        Tcl_DecrRefCount (gd -> gd_progress_cb);
    if (gd -> gd_idle_cb)
        Tcl_DecrRefCount (gd -> gd_idle_cb);

    for (entryPtr = Tcl_FirstHashEntry (&gd -> gd_keyTable, searchPtr);
            entryPtr != NULL;
            entryPtr = Tcl_NextHashEntry (searchPtr)) {
        gpgme_key_release ((GpgmeKey) Tcl_GetHashValue (entryPtr));
        Tcl_DeleteHashEntry (entryPtr);
    }
    Tcl_DeleteHashTable (&gd -> gd_keyTable);

    for (entryPtr = Tcl_FirstHashEntry (&gd -> gd_trustTable, searchPtr);
            entryPtr != NULL;
            entryPtr = Tcl_NextHashEntry (searchPtr)) {
        gpgme_trust_item_release ((GpgmeTrustItem)
                                        Tcl_GetHashValue (entryPtr));
        Tcl_DeleteHashEntry (entryPtr);
    }
    Tcl_DeleteHashTable (&gd -> gd_trustTable);

    gpgme_release (gd -> gd_context);

    Tcl_Free ((char *) data);
}


static oc_pair_t ctx_pairs[] = {
  { "info"           , aux_info            },
  { "cancel"         , ctx_cancel          },
#ifdef  notyet
  { "wait"           , ctx_wait            },
#endif
  { "get"            , ctx_getset          },     
  { "set"            , ctx_getset          }, 
  { "encrypt"        , ctx_encrypt         },
  { "decrypt"        , ctx_decrypt         },
  { "sign"           , ctx_sign            },
  { "verify"         , ctx_verify          },
  { "start-key"      , ctx_start_key       },
  { "next-key"       , ctx_next_key        },
  { "done-key"       , ctx_done_key        },
  { "info-key"       , ctx_info_key        },
  { "start-trustitem", ctx_start_trustitem },
  { "next-trustitem" , ctx_next_trustitem  },
  { "done-trustitem" , ctx_done_trustitem  },
  { "info-trustitem" , ctx_info_trustitem  },

  { NULL             , 0                   }
};

static int
ctx_new (ClientData  unused,
         Tcl_Interp *interp,
         int         objc,
         Tcl_Obj    *CONST objv[]) {
    int            result;
    char           buffer[BUFSIZ],
                  *cp;
    gpgme_ctx_t   *gd;
    Tcl_HashEntry *entryPtr;
    static int     nproc = 0;

    if (objc != 1) {
        Tcl_WrongNumArgs (interp, 1, objv, NULL);
        return TCL_ERROR;
    }

    cp = Tcl_Alloc (sizeof *gd);
    memset (cp, 0, sizeof *gd);

    gd = (gpgme_ctx_t *) cp;
    gd -> gd_dispatch = ctx_pairs;

    if ((result = gpgme_new (&gd -> gd_context)) != GPGME_No_Error) {
        Tcl_Free (cp);
        return c2t_result (interp, "gpgme_new", result);
    }
    
    sprintf (buffer, "::gpgme::ctx_%d", nproc++);
    gd -> gd_token = Tcl_CreateObjCommand (gd -> gd_interp = interp, buffer,
                                           aux_proc, (ClientData) gd,
                                           ctx_free);

    entryPtr = Tcl_CreateHashEntry (&ctxTable, (char *) gd -> gd_context,
                                    &result);
    Tcl_SetHashValue (entryPtr, (ClientData) gd);

    Tcl_InitHashTable (&gd -> gd_keyTable, TCL_STRING_KEYS);
    Tcl_InitHashTable (&gd -> gd_trustTable, TCL_STRING_KEYS);

    Tcl_SetResult (interp, buffer, TCL_VOLATILE);
    return TCL_OK;
}


/* funcions to handle recipients */


#define RECIPADD_OPERATION 0
#define RECIPADD_NAME      1
#define RECIPADD_VALIDITY  2

#define RECIPADD_MAXARGS    3
#define RECIPADD_OPTIONAL   RECIPADD_VALIDITY

static kv_pair_t recipadd_args[] = {
    { "-operation", RECIPADD_OPERATION },
    { "-name"     , RECIPADD_NAME      },
    { "-validity" , RECIPADD_VALIDITY  },

    { NULL        , 0                  }
};

static int
recip_add (ClientData  data,
           Tcl_Interp *interp,
           int         objc,
           Tcl_Obj    *CONST objv[]) {
    int            result,
                   validity;
    gpgme_recip_t *gd = (gpgme_recip_t *) data;
    Tcl_Obj       *args[RECIPADD_MAXARGS];

    memset (args, 0, sizeof *args * RECIPADD_MAXARGS);
    if ((result = crack_args (interp, objc, objv, recipadd_args,
                              RECIPADD_OPTIONAL, args)) != TCL_OK)
        return result;

    if (args[RECIPADD_VALIDITY] == NULL)
        validity = GPGME_VALIDITY_UNKNOWN;
    else {
        if ((result = tcl_GetIndexFromObjStruct (interp,
                                                 args[RECIPADD_VALIDITY],
                                                 (char **) validity_pairs,
                                                 sizeof *validity_pairs,
                                                 "validity", 0, &validity))
                != TCL_OK)
            return result;

        validity = validity_pairs[validity].p_value;
    }

    if ((result =
             gpgme_recipients_add_name_with_validity (gd -> gd_recipients,
                                                      Tcl_GetString (args[RECIPADD_NAME]),
                                                      validity))
            != GPGME_No_Error)
        return c2t_result (interp, "gpgme_add_name_with_validity", result);

    return TCL_OK;
}


static int
recip_count (ClientData  data,
             Tcl_Interp *interp,
             int         objc,
             Tcl_Obj    *CONST objv[]) {
    int            result;
    gpgme_recip_t *gd = (gpgme_recip_t *) data;

    if (objc != 3) {
        Tcl_ResetResult (interp);
        Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                          " -operation count", NULL);
        return TCL_ERROR;
    }

    result = gpgme_recipients_count (gd -> gd_recipients);

    Tcl_SetObjResult (interp, Tcl_NewLongObj (result));
    return TCL_OK;
}


static int
recip_list (ClientData  data,
            Tcl_Interp *interp,
            int         objc,
            Tcl_Obj    *CONST objv[]) {
    void          *v;
    int            result;
    const char    *cp;
    gpgme_recip_t *gd = (gpgme_recip_t *) data;
    Tcl_Obj       *resultPtr;

    if (objc != 3) {
        Tcl_ResetResult (interp);
        Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                          " -operation start-name", NULL);
        return TCL_ERROR;
    }

    if ((result = gpgme_recipients_enum_open (gd -> gd_recipients, &v))
            != GPGME_No_Error)
        return c2t_result (interp, "gpgme_recipients_enum_open", result);

    Tcl_SetObjResult (interp, resultPtr = Tcl_NewObj ());

    while ((cp = gpgme_recipients_enum_read (gd -> gd_recipients, &v)) != NULL)
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj (cp, -1));

    if ((result = gpgme_recipients_enum_close (gd -> gd_recipients, &v))
            != GPGME_No_Error)
        return c2t_result (interp, "gpgme_recipients_close", result);

    return TCL_OK;
}

/*  */

static void
recip_free (ClientData data) {
    gpgme_recip_t *gd = (gpgme_recip_t *) data;
    Tcl_HashEntry *entryPtr;

    if ((entryPtr = Tcl_FindHashEntry (&recipTable,
                                       (char *) gd -> gd_recipients)) != NULL)
        Tcl_DeleteHashEntry (entryPtr);

    if ((entryPtr = Tcl_FindHashEntry (&recipXTable, gd -> gd_name)) != NULL)
        Tcl_DeleteHashEntry (entryPtr);

    gpgme_recipients_release (gd -> gd_recipients);

    Tcl_Free ((char *) data);
}


static oc_pair_t recip_pairs[] = {
  { "info" , aux_info    }, 
  { "add"  , recip_add   },
  { "count", recip_count },
  { "list" , recip_list  },

  { NULL   , 0           }
};

static int
recip_new (ClientData  unused,
         Tcl_Interp *interp,
         int         objc,
         Tcl_Obj    *CONST objv[]) {
    int            result;
    char          *cp;
    gpgme_recip_t *gd;
    Tcl_HashEntry *entryPtr;
    static int     nproc = 0;

    if (objc != 1) {
        Tcl_WrongNumArgs (interp, 1, objv, NULL);
        return TCL_ERROR;
    }

    cp = Tcl_Alloc (sizeof *gd);
    memset (cp, 0, sizeof *gd);

    gd = (gpgme_recip_t *) cp;
    gd -> gd_dispatch = recip_pairs;

    if ((result = gpgme_recipients_new (&gd -> gd_recipients))
            != GPGME_No_Error) {
        Tcl_Free (cp);
        return c2t_result (interp, "gpgme_recipients_new", result);
    }
    
    sprintf (gd -> gd_name, "::gpgme::recip_%d", nproc++);
    gd -> gd_token = Tcl_CreateObjCommand (gd -> gd_interp = interp,
                                           gd -> gd_name, aux_proc,
                                           (ClientData) gd, recip_free);

    entryPtr = Tcl_CreateHashEntry (&recipTable, (char *) gd -> gd_recipients,
                                    &result);
    Tcl_SetHashValue (entryPtr, (ClientData) gd);

    entryPtr = Tcl_CreateHashEntry (&recipXTable, gd -> gd_name, &result);
    Tcl_SetHashValue (entryPtr, (ClientData) gd);

    Tcl_SetResult (interp, gd -> gd_name, TCL_VOLATILE);
    return TCL_OK;
}


/* miscellaneous functions */

static int
misc_check_version (ClientData  unused,
                    Tcl_Interp *interp,
                    int         objc,
                    Tcl_Obj    *CONST objv[]) {
    if (objc != 2) {
        Tcl_WrongNumArgs (interp, 1, objv, "requiredVersion");
        return TCL_ERROR;
    }
    
    Tcl_SetResult (interp,
                   (char *) gpgme_check_version (Tcl_GetString (objv[1])),
                   TCL_VOLATILE);
    return TCL_OK;
}

static int
misc_get_engine_info (ClientData  unused,
                      Tcl_Interp *interp,
                      int         objc,
                      Tcl_Obj    *CONST objv[]) {
    if (objc != 1) {
        Tcl_WrongNumArgs (interp, 1, objv, NULL);
        return TCL_ERROR;
    }

    Tcl_SetResult (interp, (char *) gpgme_get_engine_info (), TCL_VOLATILE);
    return TCL_OK;
}

static int
misc_engine_check_version (ClientData  unused,
                           Tcl_Interp *interp,
                           int         objc,
                           Tcl_Obj    *CONST objv[]) {
    int    offset,
           result;

    if (objc != 2) {
        Tcl_WrongNumArgs (interp, 1, objv, "protocol");
        return TCL_ERROR;
    }
    if ((result = tcl_GetIndexFromObjStruct (interp, objv[1],
                                             (char **) protocol_pairs,
                                             sizeof *protocol_pairs,
                                             "protocol", 0, &offset))
            != TCL_OK)
        return result;
    offset = protocol_pairs[offset].p_value;

    return c2t_result (interp, "gpgme_engine_check_version",
                       gpgme_engine_check_version (offset));
}

/*  */

#define INFO_DATATYPES  0
#define INFO_SIG_STATS  1
#define INFO_SIG_MODES  2
#define INFO_ATTRIBUTES 3
#define INFO_VALIDITY   4
#define INFO_CAPABILITY 5
#define INFO_PROTOCOLS  6
#define INFO_KEYLIST    7

static kv_pair_t info_pairs[] = {
  { "datatypes"       , INFO_DATATYPES  },
  { "signature-status", INFO_SIG_STATS  },
  { "signature-modes" , INFO_SIG_MODES  },
  { "attributes"      , INFO_ATTRIBUTES },
  { "validity"        , INFO_VALIDITY   },
  { "capabilities"    , INFO_CAPABILITY },
  { "protocols"       , INFO_PROTOCOLS  },
  { "keylist-modes"   , INFO_KEYLIST    },
   
  { NULL              , 0               }
};


static int
misc_info (ClientData  unused,
           Tcl_Interp *interp,
           int         objc,
           Tcl_Obj    *CONST objv[]) {
    int          result;
    kv_pair_t   *p;
    Tcl_Obj     *resultPtr;

    if (objc == 1)
        p = info_pairs;
    else
        if (objc != 2) {
            Tcl_ResetResult (interp);
            Tcl_AppendResult (interp, "usage: ", Tcl_GetString (objv[0]),
                              " option", NULL);
            return TCL_ERROR;
        } else {
            int     offset;
            if ((result = tcl_GetIndexFromObjStruct (interp, objv[1],
                                                     (char **) info_pairs,
                                                     sizeof *info_pairs,
                                                     "option", 0, &offset))
                    != TCL_OK)
                return result;
            offset = info_pairs[offset].p_value;

            switch (offset) {
                case INFO_DATATYPES:
                    p = dt_pairs;
                    break;
    
                case INFO_SIG_STATS:
                    p = ss_pairs;
                    break;

                case INFO_SIG_MODES:
                    p = sm_pairs;
                    break;

                case INFO_ATTRIBUTES:
                    p = attr_pairs;
                    break;
    
                case INFO_VALIDITY:
                    p = validity_pairs;
                    break;
    
                case INFO_CAPABILITY:
                    p = capability_flags;
                    break;
    
                case INFO_PROTOCOLS:
                    p = protocol_pairs;
                    break;

                case INFO_KEYLIST:
                    p = keylist_flags;
                    break;

                default:
                    Tcl_SetResult (interp,
                                   "internal error, missing known case",
                                   TCL_STATIC);
                    return TCL_ERROR;
                }
        }

    resultPtr = Tcl_NewObj ();
    for (; p -> p_key; p++)
        Tcl_ListObjAppendElement (interp, resultPtr,
                                  Tcl_NewStringObj (p -> p_key, -1));

    Tcl_SetObjResult (interp, resultPtr);
    return TCL_OK;
}



/* module initialization */

static oc_pair_t init_pairs[] = {
  { "gpgme::context"             , ctx_new                   },
#ifdef  notyet
  { "gpgme::wait"                , ctx_wait                  },
#endif
  { "gpgme::recipient"           , recip_new                 },

  { "gpgme::check_version"       , misc_check_version        },
  { "gpgme::get_engine_info"     , misc_get_engine_info      },
  { "gpgme::engine_check_version", misc_engine_check_version },

  { "gpgme::info"                , misc_info                 },

  { NULL                         , 0                         }
};


int
Tclgpgme_Init (Tcl_Interp *interp) {
    int        result;
    oc_pair_t *pp;

    if (Tcl_InitStubs (interp, "8.0", 0) == NULL)
        return TCL_ERROR;

    if (gpgme_check_version ("0.3.11") == NULL) {
      Tcl_SetResult (interp, "this package requires GPGME version >= 0.3.11",
                     TCL_STATIC);
        return TCL_ERROR;
    }
    if ((result = gpgme_engine_check_version (GPGME_PROTOCOL_OpenPGP))
            != GPGME_No_Error)
        return c2t_result (interp, "gpgme_engine_check_version", result);

    if (Tcl_PkgProvide (interp, "gpgme", VERSION) != TCL_OK)
        return TCL_ERROR;

    Tcl_InitHashTable (&ctxTable, TCL_ONE_WORD_KEYS);
    Tcl_InitHashTable (&recipTable, TCL_ONE_WORD_KEYS);
    Tcl_InitHashTable (&recipXTable, TCL_STRING_KEYS);

    gpgme_register_idle (ctx_idle_cb);

    for (pp = init_pairs; pp -> p_key; pp++)
        Tcl_CreateObjCommand (interp, pp -> p_key, pp -> p_proc,
                              (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

    return TCL_OK;
}
